from __future__ import annotations

import warnings
from collections.abc import Iterable, Sequence
from typing import Callable

from pyautoguide.references import ReferenceElement

from ._types import Direction
from .shapes import Box, BoxSpec
from .utils import line_intersects_box


class BoxArray(Sequence[Box]):
    """An immutable sequence of Box objects that proxies methods to its contents."""

    def __init__(self, boxes: Iterable[Box] | None = None) -> None:
        self._boxes = tuple(boxes) if boxes is not None else ()

    def __getitem__(self, index: int | slice) -> Box | BoxArray:
        """Returns a Box or a new BoxArray from a slice."""
        if isinstance(index, slice):
            return BoxArray(self._boxes[index])
        return self._boxes[index]

    def __len__(self) -> int:
        """Returns the number of boxes in the array."""
        return len(self._boxes)

    def __add__(self, boxes: object) -> BoxArray:
        """Extends the BoxArray with additional Box objects."""
        if not isinstance(boxes, Iterable):
            raise TypeError(f"Expected an iterable of Box objects, got {type(boxes)}")
        new_boxes = tuple(boxes)
        if not all(isinstance(box, Box) for box in new_boxes):
            raise TypeError("All items must be instances of Box.")
        return BoxArray(new_boxes + self._boxes)

    def first(self) -> Box:
        """Returns the first Box in the array."""
        if not self._boxes:
            raise IndexError("BoxArray is empty.")
        return self._boxes[0]

    def last(self) -> Box:
        """Returns the last Box in the array."""
        if not self._boxes:
            raise IndexError("BoxArray is empty.")
        return self._boxes[-1]

    def select(self, *, i: int) -> Box:
        """Returns the Box at the specified index."""
        return self._boxes[i]

    def pick(self, region: BoxSpec) -> BoxArray:
        """Returns a new BoxArray with boxes those have center inside the region."""

        new_boxes = [box for box in self._boxes if box.center in Box.from_spec(region)]
        return BoxArray(new_boxes)

    def filter_by(self, condition: Callable[[Box], bool]) -> BoxArray:
        """Returns a new BoxArray with boxes that satisfy the given condition."""
        new_boxes = [box for box in self._boxes if condition(box)]
        return BoxArray(new_boxes)

    def relative_to(
        self, direction: Direction, *, of: BoxSpec | ReferenceElement
    ) -> BoxArray:
        """Returns a new BoxArray with boxes relative to the specified direction and origin."""
        if isinstance(of, str):
            of = Box.from_spec(of)
        elif isinstance(of, ReferenceElement):
            of = of.locate().first()
        return BoxArray(
            (
                box
                for box in self._boxes
                if line_intersects_box(box, direction, of.center)
            )
        )

    def __getattr__(self, name: str):
        """Dynamically proxies method calls to the specified Box."""
        # Check if the attribute is a callable method on the Box class
        if not hasattr(Box, name) or not callable(getattr(Box, name)):
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

        def method_proxy(*args, **kwargs):
            """Proxy function that calls the method on the selected Box."""
            if not self._boxes:
                raise ValueError(f"BoxArray is empty, cannot call method {name}.")

            new_boxes = []
            for box in self._boxes:
                try:
                    return_value = getattr(box, name)(*args, **kwargs)
                    if isinstance(return_value, Box):
                        new_boxes.append(return_value)
                    elif isinstance(return_value, BoxArray):
                        new_boxes.extend(return_value._boxes)
                    else:
                        raise TypeError(
                            f"Method {name} returned unexpected type: {type(return_value)}"
                        )
                except Exception as e:
                    warnings.warn(
                        f"Method {name} raised an exception: {e}", stacklevel=2
                    )
            return BoxArray(new_boxes)

        return method_proxy



from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Self

import cv2
import numpy as np
import pyautogui as gui
from pyscreeze import Box as BoxTuple

from ._types import Direction, MouseButton
from .utils import direction_to_vector, get_search_region_in_direction

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .box_array import BoxArray

axis_pattern = re.compile(r"(?P<d>[xy]):\(?(?P<i>\d+)(?:-(?P<j>\d+))?\)?/(?P<n>\d+)")


@dataclass(frozen=True, slots=True, init=False)
class Box:
    left: int
    top: int
    width: int
    height: int

    def __init__(self, left: int, top: int, width: int, height: int):
        object.__setattr__(self, "left", int(left))
        object.__setattr__(self, "top", int(top))
        object.__setattr__(self, "width", int(width))
        object.__setattr__(self, "height", int(height))

    def to_tuple(self) -> BoxTuple:
        """Convert to a pyscreeze Box."""
        return BoxTuple(self.left, self.top, self.width, self.height)

    @classmethod
    def from_tuple(cls, box: BoxTuple) -> Box:
        """Create a Region from a pyscreeze Box."""
        return cls(left=box.left, top=box.top, width=box.width, height=box.height)

    @property
    def center(self) -> Point:
        """Get the center coordinates of the box."""
        return Point(x=self.left + self.width // 2, y=self.top + self.height // 2)

    @classmethod
    def from_spec(cls, spec: BoxSpec, shape: tuple[int, int] | None = None) -> Box:
        if isinstance(spec, Box):
            return spec
        if shape is None:
            img = np.array(gui.screenshot())
            shape = (img.shape[0]), (img.shape[1])

        default_box = {"left": 0, "top": 0, "width": shape[1], "height": shape[0]}

        axis_mapping = {"x": ("left", "width", 1), "y": ("top", "height", 0)}
        for axis, i, j, n in axis_pattern.findall(spec):
            alignment, size_attr, dim_index = axis_mapping[axis]
            size = shape[dim_index] // int(n)
            i, j = int(i), int(j) if j else int(i)
            default_box.update({
                alignment: (i - 1) * size,
                size_attr: (j - i + 1) * size,
            })

        return cls(**default_box)

    def log_screenshot(self, filename: str | Path):
        """Take a screenshot of the box and save it to a file."""
        img = gui.screenshot(region=self.to_tuple())
        img.save(filename)
        logger.info(f"Screenshot saved to {filename}")
        return self

    def resolve(self, base: BoxSpec | None) -> Box:
        if base is None:
            return self
        if isinstance(base, str):
            base = Box.from_spec(base)
        return Box(
            left=self.left + base.left,
            top=self.top + base.top,
            width=self.width,
            height=self.height,
        )

    def click(self, clicks: int = 1, button: MouseButton = "left") -> Self:
        """Click at the center of the box with optional offset."""
        from .actions import move_and_click

        move_and_click(target=self, clicks=clicks, button=button)
        return self

    def offset(self, direction: Direction, shift: int = 0) -> Box:
        """Return a new Box offset in the specified direction."""
        vector = direction_to_vector(direction)
        return Box(
            left=self.left + vector[0] * shift,
            top=self.top + vector[1] * shift,
            width=self.width,
            height=self.height,
        )

    def __contains__(self, point: Point) -> bool:
        """Check if a Point is inside the Box."""
        return (
            self.left <= point.x <= self.left + self.width
            and self.top <= point.y <= self.top + self.height
        )

    def intersect(self, other: BoxSpec) -> Box:
        """Return the intersection of two Boxes."""
        if not isinstance(other, Box):
            other = Box.from_spec(other)
        left = max(self.left, other.left)
        top = max(self.top, other.top)
        right = min(self.left + self.width, other.left + other.width)
        bottom = min(self.top + self.height, other.top + other.height)

        if left < right and top < bottom:
            return Box(left=left, top=top, width=right - left, height=bottom - top)
        raise ValueError("Boxes do not intersect.")

    def find_color(
        self,
        color: tuple[int, int, int],
        towards: Direction,
        *,
        region: BoxSpec | None = None,
        tolerance: int | None = None,
    ) -> BoxArray:
        """Find the median pixel with given color in the direction."""
        from .box_array import BoxArray

        # Determine search region based on direction
        assert isinstance(towards, str), "integer direction is not supported"
        search_region = get_search_region_in_direction(self, towards, size=gui.size())

        # Apply optional region constraint
        if region:
            given_region = Box.from_spec(region)
            search_region = search_region.intersect(given_region)

        # Take screenshot of search region
        img = gui.screenshot(region=search_region.to_tuple())

        # Find connected components of the target color
        color_array = np.array(color)
        if tolerance is None:
            img_color_mask = np.all(np.array(img) == color_array, axis=-1).astype(
                np.uint8
            )
        else:
            img_color_mask = np.all(
                np.abs(np.array(img) - color_array) <= tolerance, axis=-1
            ).astype(np.uint8)
        n, img_label = cv2.connectedComponents(
            img_color_mask, connectivity=8, ltype=cv2.CV_32S
        )

        # Find all bounding boxes of connected components
        det_boxes = []
        for i in range(1, n):
            mask = img_label == i
            coords = np.column_stack(np.where(mask))
            if coords.size == 0:
                continue
            y_min, x_min = coords.min(axis=0)
            y_max, x_max = coords.max(axis=0)
            # Convert to absolute coordinates
            det_boxes.append(
                Box(
                    left=x_min,
                    top=y_min,
                    width=x_max - x_min + 1,
                    height=y_max - y_min + 1,
                )
            )

        if not det_boxes:
            raise ValueError(
                f"No pixels with color {color} found in direction {towards}"
            )

        return BoxArray([b.resolve(search_region) for b in det_boxes])


type BoxSpec = Box | str


@dataclass(frozen=True, slots=True)
class Point:
    x: int
    y: int

    def to_tuple(self) -> tuple[int, int]:
        """Convert to a tuple."""
        return (self.x, self.y)

    def __add__(self, other: Point | np.ndarray) -> Point:
        """Add another Point or a numpy array to this Point."""
        if isinstance(other, Point):
            return Point(self.x + other.x, self.y + other.y)
        elif isinstance(other, np.ndarray):
            return Point(self.x + int(other[0]), self.y + int(other[1]))
        raise TypeError("Unsupported type for addition with Point.")

    def __radd__(self, other: Point | np.ndarray) -> Point:
        """Add this Point to another Point or a numpy array."""
        return self.__add__(other)

    def __iter__(self):
        """Return an iterator over the Point coordinates."""
        yield self.x
        yield self.y

    @classmethod
    def from_tuple(cls, point: tuple[int, int]) -> Point:
        """Create a Point from a tuple."""
        return cls(x=point[0], y=point[1])

    def __array__(self, dtype=None, copy=None):
        return np.array([self.x, self.y], dtype=dtype, copy=copy)

    def resolve(self, base: Point | None) -> Point:
        """Return a new Point offset by another Point (base)."""
        if base is None:
            return self
        if isinstance(base, tuple):
            base = Point.from_tuple(base)
        return Point(self.x + base.x, self.y + base.y)

    def offset(self, direction: Direction, offset: int = 0) -> Point:
        """Return a new Point offset in the specified direction."""
        vector = direction_to_vector(direction)
        return self + vector * offset

    def click(self, clicks: int = 1, button: MouseButton = "left"):
        """Click at the Point location."""

        from .actions import move_and_click

        move_and_click(target=self, clicks=clicks, button=button)
        return self



from __future__ import annotations

import logging
from keyword import iskeyword
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import networkx as nx
import numpy as np
import pydot
from transitions.extensions import GraphMachine

if TYPE_CHECKING:
    from .shapes import Box, Point

from ._types import Direction

logger = logging.getLogger(__name__)


def is_valid_variable_name(name: str) -> bool:
    return name.isidentifier() and not iskeyword(name)


def get_file(
    dir: Path, *, name: str, file_type: Literal["image", "text"] = "image"
) -> Path:
    """Find a file in the given directory."""
    image_file_extensions = [".png", ".jpg", ".jpeg", ".bmp"]
    text_file_extensions = [".txt", ".md", ".json"]

    if file_type == "image":
        valid_extensions = image_file_extensions
    elif file_type == "text":
        valid_extensions = text_file_extensions
    else:
        raise ValueError(f"Unknown file type: {file_type}")

    for path in dir.glob(f"{name}.*"):
        if path.suffix in valid_extensions:
            return path
    raise FileNotFoundError(f"File {name} not found in directory {dir}")


def get_nx_graph(machine: GraphMachine) -> nx.MultiDiGraph:
    pydot_graph = pydot.graph_from_dot_data(machine.get_graph().source)[0]  # type: ignore
    nx_graph = nx.nx_pydot.from_pydot(pydot_graph)
    return nx_graph


def direction_to_vector(direction: Direction) -> np.ndarray:
    mapping = {
        "right": 0,
        "top": 90,
        "left": 180,
        "bottom": 270,
        "top-left": 135,
        "top-right": 45,
        "bottom-left": 225,
        "bottom-right": 315,
    }
    deg = mapping[direction] if isinstance(direction, str) else direction
    rad = np.deg2rad(deg)
    return np.array([np.cos(rad), -np.sin(rad)])


def line_intersects_box(
    box: Box, direction: Direction, point: Point
) -> tuple[Point, ...]:
    """Check if a box intersects with a line in a specific direction."""
    d = direction_to_vector(direction)
    sides = [
        ((box.left, box.top), (0, -1), box.height),  # left side
        ((box.left, box.top + box.height), (1, 0), box.width),  # bottom side
        (
            (box.left + box.width, box.top + box.height),
            (0, 1),
            box.height,
        ),  # right side
        ((box.left + box.width, box.top), (-1, 0), box.width),  # top side
    ]
    intersections = ()
    for corner, line_dir, length in sides:
        line_start = np.array(corner)
        line_dir_ = np.array(line_dir)
        try:
            t, u = np.linalg.solve(np.array([line_dir_, -d]).T, point - line_start)
        except np.linalg.LinAlgError:
            continue
        if 0 <= t <= length and u >= 0:
            intersections += (point + (u * d),)
    return intersections


def get_search_region_in_direction(
    box: Box, towards: Direction, size: tuple[int, int]
) -> Box:
    """Get the search region in a specific direction from the given box."""
    from .shapes import Box

    if towards == "top-right":
        return Box(
            left=box.left + box.width,
            top=0,
            width=size[0] - (box.left + box.width),
            height=box.top,
        )
    elif towards == "top-left":
        return Box(left=0, top=0, width=box.left, height=box.top)
    elif towards == "bottom-right":
        return Box(
            left=box.left + box.width,
            top=box.top + box.height,
            width=size[0] - (box.left + box.width),
            height=size[1] - (box.top + box.height),
        )
    elif towards == "bottom-left":
        return Box(
            left=0,
            top=box.top + box.height,
            width=box.left + box.width,
            height=size[1] - (box.top + box.height),
        )
    elif towards == "top":
        return Box(left=box.left, top=0, width=box.width, height=box.top)
    elif towards == "bottom":
        return Box(
            left=box.left,
            top=box.top + box.height,
            width=box.width,
            height=size[1] - (box.top + box.height),
        )
    elif towards == "left":
        return Box(left=0, top=box.top, width=box.left, height=box.height)
    elif towards == "right":
        return Box(
            left=box.left + box.width,
            top=box.top,
            width=size[0] - (box.left + box.width),
            height=box.height,
        )
    else:
        raise ValueError(f"Unknown direction: {towards}")


# def generate_points(
#     source: Point, towards: Direction, offset: int = 0, max_distance: int = 10_000_000
# ):
#     step = direction_to_vector(towards)
#     i = offset
#     while True:
#         yield Point(*np.round(source + step * i))
#         i += 1
#         if i > max_distance:
#             break
